import matplotlib.pyplot as plt
from numpy import ndarray
from typing import Union
from matplotlib.colors import Normalize, Colormap, ListedColormap, LinearSegmentedColormap

CONTINUOUS = "continuous"
DISCRETE = "discrete"
HORIZONTAL = "horizontal"
VERTICAL = "vertical "

def get_norm(values: ndarray, vmin: float, vmax: float) -> Normalize:
    """Get `Normalize` by the provided `vmin` and `vmax`

    Parameters
    ----------
    values : ndarray
        values are used to normalize
    vmin : float
        the minemum value visualized
    vmax : float
        the maximum value visualized
    Returns
    -------
    Normalize
    """
    vmin = vmin if vmin is not None else values.min()
    vmax = vmax if vmax is not None else values.max()
    return Normalize(vmin=vmin, vmax=vmax)


def cycle_cmap(cmap: Colormap, num: int) -> Colormap:
    """When the number of discrete colors are not enough for categories, cycle the colors of cmap 
    to meet the number of categories.

    Parameters
    ----------
    cmap : Colormap
        Colormap of categories
    num : int
        the number of categories

    Returns
    -------
    Colormap
    """
    colors = cmap.colors * (num // cmap.N) + cmap.colors[:(num % cmap.N)]
    return ListedColormap(colors)


def get_cmap(cmap: Union[Colormap, str, list], cmap_type: str = CONTINUOUS) -> Colormap:
    """Transform different color expresion types to Colormap

    Parameters
    ----------
    cmap : Union[Colormap, str, list]
        the raw color expression
    cmap_type : str, optional
        cmap is used for bar type, by default CONTINUOUS

    Returns
    -------
    Colormap
    """
    if isinstance(cmap, str):
        cmap =  plt.colormaps[cmap]
    elif isinstance(cmap, list):
        if cmap_type == CONTINUOUS:
            cmap = LinearSegmentedColormap.from_list("from_list", colors=cmap)
        else:
            cmap = ListedColormap(colors=cmap)
    elif isinstance(cmap, Colormap):
        pass
    else:
        raise TypeError("'cmap' must be `Colormap` type!")
    
    if (cmap_type == CONTINUOUS) and (not isinstance(cmap, LinearSegmentedColormap)):
        cmap = LinearSegmentedColormap.from_list("from_list", cmap.colors)
        return cmap
    elif (cmap_type == DISCRETE) and (not isinstance(cmap, ListedColormap)):
        raise TypeError(f"'cmap' must be 'ListedColormap' for {DISCRETE}")
    else:
        return cmap



import numpy as np
import matplotlib.pyplot as plt
from typing import Union
from numpy import ndarray
from matplotlib.colors import Colormap
from matplotlib.axes import Axes
from ._utils import get_norm, get_cmap


class Heatmap:
    def __init__(
        self, mat: ndarray,
        cmap: Union[str, Colormap, list], vmin: float = None, vmax: float = None,
        name: str = None, rownames: ndarray = None, colnames: ndarray = None,
        rownames_side: str = "left", colnames_side: str = "top",
        rownames_style: dict = dict(rotation=0), colnames_style: dict = dict(rotation=0),
        edgecolor: str = "none", edgewidth: float = 1
    ) -> None:
        """Heatmap

        Parameters
        ----------
        mat : ndarray
            the matrix used to plot heatmap
        cmap : Union[str, Colormap, list]
            colormap for the matrix
        vmin : float, optional
            the minemum value scaled. by default None, use the minemum value of matrix. 
        vmax : float, optional
            the maximum value scaled. by default None, use the maximum value of matrix.
        name : str, optional
            heatmap name, show it on its legend. by default None, don't show the name
        rownames : ndarray, optional
            row names of heatmap. by default None, don't show the row names
        colnames : ndarray, optional
            column names of heatmap. by default None, don't show the column names
        rownames_side : str, optional
            show row names on which side("left" or "right")? by default "left"
        colnames_side : str, optional
            show column names on which side("top" or "bottom")?, by default "top"
        rownames_style : dict, optional
            modify row names' style, such as, fontsize, fontstyle, etc. See more information on 
            [`matplotlib.text.Text`](https://matplotlib.org/stable/api/text_api.html#matplotlib.text.Text).
            by default dict(rotation=0)
        colnames_style : dict, optional
            See `rownames_style`, by default dict(rotation=0)
        edgecolor : str, optional
            the color of heatmap's cell edge, by default "none", no edge. 
            !Note: If provide `None`, will use the `rcParams["patch.edgecolor"]`, 
            it default as "black". 
        edgewidth : float, optional
            the width of heatmap's cell edge, by default 1
        """
        self.mat = mat
        self.name = name

        self.cmap = get_cmap(cmap)
        self.norm = get_norm(self.mat, vmin, vmax)

        self.nrows, self.ncols = self._get_nrows_ncols()
        self.rownames = self._check_names(axis="row", names=rownames)
        self.colnames = self._check_names(axis="col", names=colnames)
        self.sides = self._parse_name_side(rownames_side, colnames_side)

        self.rownames_style, self.colnames_style = rownames_style, colnames_style
        self.edgecolor = edgecolor
        self.edgewidth = edgewidth

    def _get_nrows_ncols(self):
        return self.mat.shape

    def _parse_name_side(self, rownames_side: str, colnames_side: str) -> dict:
        """Parse the row/column names' side

        Parameters
        ----------
        rownames_side : str
            show row names on which side("left" or "right")? by default "left"
        colnames_side : str
            show column names on which side("top" or "bottom")?, by default "top"

        Returns
        -------
        dict
            used in `tick_params`, will change row/column names side

        Raises
        ------
        KeyError
            If the rownames_side/colnames_side provided are not correct, will raise KeyError
        """
        side_dict = {"labeltop": False, "labelbottom": False,
                     "labelleft": False, "labelright": False}

        # Check row/colnames_side options
        rownames_side_options = ["left", "right"]
        colnames_side_options = ["top", "bottom"]
        if rownames_side not in rownames_side_options:
            raise KeyError(
                f"The rownames_side, '{rownames_side}' is not one of {rownames_side_options}")
        if colnames_side not in colnames_side_options:
            raise KeyError(
                f"The colnames_side, '{colnames_side}' is not one of {colnames_side_options}")

        side_dict[f"label{rownames_side}"] = all([rownames_side, self.rownames is not None])
        side_dict[f"label{colnames_side}"] = all([colnames_side, self.colnames is not None])
        return side_dict

    def _check_names(self, axis: str, names: ndarray) -> ndarray:
        """Check the row/column names provided are correct

        Parameters
        ----------
        axis : str
            "row" or "col"?
        names : ndarray
            the row/column names provided

        Returns
        -------
        ndarray

        Raises
        ------
        ValueError
            If the row/column names provided are not correct will raise KeyError
        """
        num = self.nrows if axis == "row" else self.ncols
        if names is None:
            return names
        elif len(names) != num:
            raise ValueError(
                f"The length of {axis}names is not equal to the number of heatmap matrix!")
        else:
            return names

    def draw(self, ax: Axes) -> None:
        ax.imshow(self.mat, norm=self.norm, cmap=self.cmap, aspect="auto")

        # Set row/colnames and their font style(rotation, family, size, etc)
        ax.set_xticks(np.arange(self.ncols), labels=self.colnames,
                      minor=False, **self.colnames_style)
        ax.set_yticks(np.arange(self.nrows), labels=self.rownames,
                      minor=False, **self.rownames_style)

        # Set ticks and ticklabels location and if show them
        ax.tick_params(
            axis="both", which="major", pad=0, top=False, bottom=False, left=False, right=False,
            **self.sides
        )

        # Configure edges color and width
        if self.edgecolor:
            ax.set_xticks(np.arange(-0.5, self.ncols), minor=True)
            ax.set_yticks(np.arange(-0.5, self.nrows), minor=True)
            ax.tick_params(axis="both", which="minor", pad=0,
                           top=False, bottom=False, left=False, right=False)
            # For whole Axes
            ax.spines[:].set_color(self.edgecolor)
            ax.spines[:].set_linewidth(self.edgewidth)
            # For every cells
            ax.grid(True, axis="both", which="minor",
                    color=self.edgecolor, linewidth=self.edgewidth)
        else:
            ax.spines[:].set_visible(False)
            ax.grid(False, axis="both", which="both")



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from numpy import ndarray, number
from pandas import DataFrame, Series
from typing import Union, Dict, Tuple, List
from matplotlib.colors import Colormap, Normalize, BoundaryNorm
from matplotlib.axes import Axes
from ._utils import get_norm, get_cmap, cycle_cmap, CONTINUOUS, DISCRETE, HORIZONTAL, VERTICAL


def _object2categrey(anno: DataFrame) -> DataFrame:
    """Transform columns without `category` or `numpy.number` type to `category`"""
    need_transform_columns = anno.select_dtypes(exclude=[np.number, "category"]).columns
    return anno.apply(lambda x: x.astype("category") if x.name in need_transform_columns else x,
                      axis=0)


def _get_bartype(values: Series) -> str:
    if values.dtype in ["int", "float"]:
        return CONTINUOUS
    else:
        return DISCRETE


def _transform_discrete_values(values: Series) -> Tuple[ndarray, Dict[str, number]]:
    """Deal with discrete values, transform categories to numbers and create mapper for categories 
    and their numbers

    Parameters
    ----------
    values : Series
        categories
    Returns
    -------
    Tuple[ndarray, Dict[str, number]]
        return transformed numeric values and categories-numbers mapper
    """
    values_order = values.dtype.categories.to_list()
    values_mapper = {k: v for v, k in enumerate(values_order)}
    values = values.map(values_mapper)
    return values, values_mapper


class AnnotationBar:
    def __init__(
        self,
        values: ndarray, cmap: Union[str, Colormap, List],
        values_mapper: Dict[str, number] = None,
        name: str = None, vmin: float = None, vmax: float = None,
        bartype: str = CONTINUOUS, direction: str = HORIZONTAL,
        tick_labels_params: Dict = dict(size=6)
    ) -> None:
        """single AnnotationBar

        Parameters
        ----------
        values : ndarray
            values are used to draw annotation bar
        cmap : Union[str, Colormap, List]
            colormap for annotation bar
        values_mapper : Dict[str, number], optional
            values_mapper is used to create legend for DISCRETE values, by default None
        name : str, optional
            AnnotationBar name, show it on the bar side and its legend, by default None
        vmin : float, optional
            the minemum value scaled. `None` means use the minemum value of values. by default None
        vmax : float, optional
            the maximum value scaled. `None` means use the maximum value of values. by default None
        bartype : str, optional
            bar values are CONTINUOUS or DISCRETE, by default CONTINUOUS
        direction : str, optional
            visualize bar as HORIZONTAL or VERTICAL, by default HORIZONTAL
        """
        self.name = name
        self.direction = direction
        self.bartype = self._check_bartype(bartype)
        self.name_attrs = self._get_name_attrs()

        self.values = self._check_direction(values)
        self.values_mapper = values_mapper
        self.cmap = get_cmap(cmap, self.bartype)
        self.norm = self._get_norm(vmin, vmax)
        self.tick_labels_params = tick_labels_params

    def _check_bartype(self, bartype: str) -> str:
        """Validate `bar_type`"""
        if bartype in [CONTINUOUS, DISCRETE]:
            return bartype
        else:
            raise KeyError(f"`bar_type` have to be chose from {[CONTINUOUS, DISCRETE]}!")

    def _check_direction(self, values: ndarray) -> ndarray:
        """Validate `direction` and reshape `values`"""
        if self.direction == HORIZONTAL:
            return values.reshape(1, -1)
        elif self.direction == VERTICAL:
            return values.reshape(-1, 1)
        else:
            raise KeyError(f"`direction` have to be chose from {[HORIZONTAL, VERTICAL]}!")

    def _get_norm(self, vmin: float, vmax: float) -> Normalize:
        """`Normalize` for CONTINUOUS and `BoundaryNorm` for DISCRETE

        Parameters
        ----------
        vmin : float
            the minemum value scaled
        vmax : float
            the maximum value scaled

        Returns
        -------
        Normalize
        """
        if self.bartype == CONTINUOUS:
            return get_norm(self.values, vmin, vmax)
        else:
            num_colors = len(self.values_mapper)
            # Update cmap
            self.cmap = cycle_cmap(self.cmap, num_colors)
            bounds = np.arange(-0.5, num_colors)
            return BoundaryNorm(bounds, num_colors)

    def _get_name_attrs(self) -> Dict:
        """Config name's attrs used in visualized

        Returns
        -------
        Dict
            Get more information, see: 
            [matplotlib.axes.Axes.tick_params](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.tick_params.html)
        """
        side_dict = {"labeltop": False, "labelbottom": False,
                     "labelleft": False, "labelright": False}
        if self.direction == HORIZONTAL:
            side_dict["labelright"] = bool(self.name)
            side_dict["labelrotation"] = 0
        else:
            side_dict["labelbottom"] = bool(self.name)
            side_dict["labelrotation"] = 90
        return side_dict

    def draw(self, ax: Axes) -> None:
        ax.imshow(self.values, norm=self.norm, cmap=self.cmap, aspect="auto")
        ax.tick_params(
            axis="both", pad=0, top=False, bottom=False, left=False, right=False,
            **self.name_attrs
        )
        if self.direction == HORIZONTAL:
            ax.set_yticks([0], [self.name], **self.tick_labels_params)
        elif self.direction == VERTICAL:
            ax.set_xticks([0], [self.name], **self.tick_labels_params)
        else:
            raise KeyError("`direction` have to be chose from ['horizontal', 'vertical']!")
        ax.spines[:].set_visible(False)
        ax.grid(False)


class ListAnnotationBar:
    def __init__(
        self, anno: DataFrame, cmaps: Dict[str, Union[str, Colormap, List]],
        direction: str = HORIZONTAL, show_names: bool = True,
        tick_labels_params: Dict = dict(size=6)
    ) -> None:
        """Contain multiple Annotationbars

        Parameters
        ----------
        anno : DataFrame
            DataFrame used to create Annotationbar
        cmaps : Dict[str, Union[str, Colormap, List]]
            Colormaps for each Annotationbar, keys are the DataFrame's columns
        direction : str, optional
            visualize bar as HORIZONTAL or VERTICAL, by default HORIZONTAL
        show_names : bool, optional
            show AnnotationBar name or not, by default True
        tick_labels_params: Dict
        """
        anno = _object2categrey(anno)

        self.cmaps = cmaps
        self.direction = direction
        self.show_names = show_names
        self.tick_labels_params = tick_labels_params
        self.annotationbars = self._get_annotation_bars(anno)

    def _get_annotation_bars(self, anno: DataFrame) -> List[AnnotationBar]:
        """Create AnnotationBars along columns

        Parameters
        ----------
        anno : DataFrame
            DataFrame used to create Annotationbar

        Returns
        -------
        List[AnnotationBar]
        """
        annotationbars = []
        for name, values in anno.items():
            bartype = _get_bartype(values)
            if bartype == CONTINUOUS:
                cmap = self.cmaps.pop(name, "viridis")
                values_mapper = None
            else:
                values, values_mapper = _transform_discrete_values(values)
                cmap = self.cmaps.pop(name, "tab20")
            name = name if self.show_names else None
            tmp_annobar = AnnotationBar(
                values=values.to_numpy(), cmap=cmap, values_mapper=values_mapper,
                name=name, vmin=None, vmax=None, bartype=bartype, direction=self.direction,
                tick_labels_params=self.tick_labels_params
            )
            annotationbars.append(tmp_annobar)
        return annotationbars

    def draw(self, axes: List[Axes]) -> None:
        for annobar, ax in zip(self.annotationbars, axes):
            annobar.draw(ax)



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from typing import Sequence, Dict
from matplotlib.colors import Colormap, Normalize, BoundaryNorm
from matplotlib.axes import Axes
from ._utils import CONTINUOUS, DISCRETE


class Legend:
    def __init__(
        self, cmap: Colormap, norm: Normalize,
        tick_locs: Sequence, tick_labels: Sequence,
        name: str = None, tick_labels_params: Dict = dict(size=8),
        title_params: Dict = dict(size=10),
        bartype: str = CONTINUOUS
    ) -> None:
        """_summary_

        Parameters
        ----------
        cmap : Colormap
            the colormap which is the same as its owner
        norm : Normalize
            the Normalize which is the same as its owner
        tick_locs : Sequence, optional
            the tick locations
        tick_labels : Sequence, optional
            the tick labels, which must be match with tick locations
        name : str, optional
            the legend name get from its owner, by default None, don't show it
        tick_labels_params : Dict, optional
            modify tick labels' style, by default dict(size=8)
        title_params : Dict, optional
            modify title's style, by default dict(size=10)
        bartype : str, optional
            the legend bar type(CONTINUOUS or DISCRETE), by default CONTINUOUS
        """
        self.ticks = np.array(tick_locs)
        self.labels = np.array(tick_labels)
        self.tick_labels_params = tick_labels_params
        self.title_params = title_params
        self.cmap = cmap
        self.norm = self._scale_norm(norm)
        self.name = name if name else ""
        self.bartype = bartype
        self.cbar = None
    
    def _scale_norm(self, norm):
        """Scale the continuous norm by the tick_locs"""
        if isinstance(norm, BoundaryNorm):
            return norm
        else:
            if self.ticks.min() < norm.vmin:
                vmin = self.ticks.min()
            else:
                vmin = norm.vmin
            if self.ticks.max() > norm.vmax:
                vmax = self.ticks.max()
            else:
                vmax = norm.vmax
            return Normalize(vmin, vmax)

    def draw(self, ax: Axes) -> None:
        self.cbar = mpl.colorbar.Colorbar(
            ax, mpl.cm.ScalarMappable(norm=self.norm, cmap=self.cmap),
            orientation="vertical", drawedges=False, filled=True
        )
        ax.set_yticks(self.ticks, self.labels, **self.tick_labels_params)
        ax.invert_yaxis()
        ax.set_title(self.name, loc="left", **self.title_params)
        ax.spines[:].set_visible(False)
        ax.tick_params(
            axis="both", pad=0, top=False, bottom=False, left=False, right=False,
            labeltop=False, labelbottom=False, labelleft=False, labelright=True
        )
        if self.bartype == CONTINUOUS:
            ax.tick_params(axis="y", pad=2, direction="in", color="#DADADA", left=True, right=True)
        elif self.bartype == DISCRETE:
            pass
        else:
            raise KeyError(f"'bartype' must be {[CONTINUOUS, DISCRETE]}!")



import matplotlib.pyplot as plt
from typing import List, Sequence, Union
from matplotlib.axes import Axes
from matplotlib.gridspec import GridSpecFromSubplotSpec, SubplotSpec, GridSpecBase


class Layout:
    def __init__(
        self,
        center_width: float, center_height: float,
        left_width: float, top_height: float,
        right_width: float, bottom_height: float,
        sub_left_width: Sequence[float], sub_top_height: Sequence[float],
        sub_right_width: Sequence[float], sub_bottom_height: Sequence[float],
        wspace: float, hspace: float,
        sub_left_wspace: float, sub_top_hspace: float,
        sub_right_wspace: float, sub_bottom_hspace: float,
        width: float = None, height: float = None
    ) -> None:
        """_summary_

        Parameters
        ----------
        center_width : float
            the center region's relative width
        center_height : float
            the center region's relative height
        left_width : float
            the left region's relative width
        top_height : float
            the top region's relative height
        right_width : float
            the right region's relative width
        bottom_height : float
            the bottom region's relative height
        sub_left_width : Sequence[float]
            the sub region's ratio in the left region
        sub_top_height : Sequence[float]
            the sub region's ratio in the top region
        sub_right_width : Sequence[float]
            the sub region's ratio in the right region
        sub_bottom_height : Sequence[float]
            the sub region's ratio in the bottom region
        wspace : float
            the space between regions at width direction
        hspace : float
            the space between regions at height direction
        sub_left_wspace : float
            the space between left sub regions at width direction
        sub_top_hspace : float
            the space between top sub regions at height direction
        sub_right_wspace : float
            the space between right sub regions at width direction
        sub_bottom_hspace : float
            the space between bottom sub regions at height direction
        width : float, optional
            the real width of whole figure, by default None, use the sum of regions' relative width
        height : float, optional
            the real height of whole figure, by default None, use the sum of regions' relative height
        """
        self.width = center_width + left_width + right_width if width is None else width
        self.height = center_height + top_height + bottom_height if height is None else height

        self.center_w, self.center_h = center_width, center_height
        self.left_w, self.top_h = left_width, top_height
        self.right_w, self.bottom_h = right_width, bottom_height

        self.sub_left_w, self.sub_top_h = sub_left_width, sub_top_height
        self.sub_right_w, self.sub_bottom_h = sub_right_width, sub_bottom_height

        self.wspace, self.hspace = wspace, hspace
        self.sub_left_wspace, self.sub_top_hspace = sub_left_wspace, sub_top_hspace
        self.sub_right_wspace, self.sub_bottom_hspace = sub_right_wspace, sub_bottom_hspace

        self.fig, self.gs = self._create_gridspec()
        self.left_gs = self._create_subgridspec(
            self.gs[1, 0], None, self.sub_left_wspace, [1], self.sub_left_w)
        self.right_gs = self._create_subgridspec(
            self.gs[1, 2], None, self.sub_right_wspace, [1], self.sub_right_w)
        self.top_gs = self._create_subgridspec(
            self.gs[0, 1], self.sub_top_hspace, None, self.sub_top_h, [1])
        self.bottom_gs = self._create_subgridspec(
            self.gs[2, 1], self.sub_bottom_hspace, None, self.sub_bottom_h, [1])

    def _create_gridspec(self):
        fig = plt.figure(figsize=(self.width, self.height))
        gs = fig.add_gridspec(
            nrows=3, ncols=3,
            hspace=self.hspace, wspace=self.wspace,
            height_ratios=[self.top_h, self.center_h, self.bottom_h],
            width_ratios=[self.left_w, self.center_w, self.right_w]
        )
        return fig, gs

    def _create_subgridspec(
        self, sps: SubplotSpec, hspace: float, wspace: float,
        h_ratios: Sequence, w_ratios: Sequence
    ) -> GridSpecFromSubplotSpec:
        """Create the sub `GridSpec`

        Parameters
        ----------
        sps : SubplotSpec
            the `SubplotSpec` generated from main `GridSpec`
        hspace : float
            the space between `SubplotSpec`s from the sub `GridSpec` at height direction 
        wspace : float
            the space between `SubplotSpec`s from the sub `GridSpec` at width direction
        h_ratios : Sequence
            the ratios of heights of `SubplotSpec`s  from the sub `GridSpec`
        w_ratios : Sequence
            the ratios of widths of `SubplotSpec`s  from the sub `GridSpec`

        Returns
        -------
        GridSpecFromSubplotSpec
        """
        gs_sps = sps.subgridspec(
            nrows=len(h_ratios), ncols=len(w_ratios),
            hspace=hspace, wspace=wspace,
            height_ratios=h_ratios, width_ratios=w_ratios
        )
        return gs_sps

    def create_axes(
        self, sps: Union[GridSpecBase, SubplotSpec], axis: int = 0
    ) -> Union[Axes, List[Axes]]:
        """Create List Axes or single from `GridSpec` or `SubplotSpec` 

        Parameters
        ----------
        sps : Union[GridSpecBase, SubplotSpec]
        axis : int, optional
            create Axes along which axis, by default 0, along the row

        Returns
        -------
        Union[Axes, List[Axes]]
        """
        if isinstance(sps, SubplotSpec):
            return self.fig.add_subplot(sps)
        if isinstance(sps, GridSpecBase):
            axes = []
            if axis == 0:
                for i in range(sps.nrows):
                    axes.append(self.fig.add_subplot(sps[i, 0]))
                return axes
            elif axis == 1:
                for i in range(sps.ncols):
                    axes.append(self.fig.add_subplot(sps[0, i]))
                return axes
            else:
                raise KeyError("'axis' must be '0' or '1'!")



import numpy as np
import pandas as pd
from pandas import DataFrame
from numpy import ndarray
from typing import Union, Sequence, Dict
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from ._heatmap import Heatmap
from ._annotation import ListAnnotationBar
from ._legend import Legend
from ._layout import Layout
from ._utils import HORIZONTAL, VERTICAL, CONTINUOUS


def none2dict(x: Dict = None) -> Dict:
    """Transform `None` to null `Dict`"""
    if x is None:
        return dict()
    else:
        return x


def check_margin_names(df_margin_names: ndarray, margin_names: Sequence = None,
                       show_margin_names: bool = True, axis: str = "row") -> Union[ndarray, None]:
    """Check row/column names are correct

    Parameters
    ----------
    df_margin_names : ndarray
        the main DataFrame's row/column names
    margin_names : Sequence, optional
        the row/column names provided, by default None
    show_margin_names : bool, optional
        whether show row/column names, by default True
    axis : str, optional
        "row" or "col"? by default "row"

    Returns
    -------
    Union[ndarray, None]

    Raises
    ------
    ValueError
        If provide row/column names and its length is not match the main DataFrame's, will raise
        ValueError
    """
    if not show_margin_names:
        return None
    elif margin_names is None:
        return df_margin_names
    elif len(df_margin_names) == len(margin_names):
        return np.array(margin_names)
    else:
        raise ValueError(f"The length of {axis}names is not match `mat`!")


def create_annotation(
        anno: Union[DataFrame, None], cmaps: Dict[str, Union[str, Colormap, list]],
        names_style: Dict, show_names: bool, expected_nrows: int, axis="row"
) -> Union[ListAnnotationBar, None]:
    """Instance row/column `ListAnnotationBar`

    Parameters
    ----------
    anno : Union[DataFrame, None]
        the annotation's DataFrame
    cmaps : Dict[str, Union[str, Colormap, list]]
        the colormaps for each AnnotationBar
    show_names : bool
        whether show AnnotationBar's name
    expected_nrows : int
        the number of row/columns of main DataFrame
    axis : str, optional
        "row" or "col" annotation?, by default "row"
    names_style: Dict[str, Dict], optional
        modify the style of row AnnotationBars' label.


    Returns
    -------
    Union[ListAnnotationBar, None]

    Raises
    ------
    ValueError
        If the number of rows of annotation's DataFrame is not match the main DataFrame's row/column
        number, will raise ValueError
    """
    if anno is None:
        return None
    elif anno.shape[0] == expected_nrows:
        axis = VERTICAL if axis == "row" else HORIZONTAL
        cmaps = none2dict(cmaps)
        return ListAnnotationBar(
            anno=anno, cmaps=cmaps, direction=axis, show_names=show_names,
            tick_labels_params=names_style
        )
    else:
        raise ValueError(f"The number of annotation_{axis}'s rows is not match `mat`!")


def pheatmap(
    mat: DataFrame,
    cmap: Union[str, Colormap, list] = "bwr",
    vmin: float = None, vmax: float = None,
    name: str = None, rownames: ndarray = None, colnames: ndarray = None,
    rownames_side: str = "left", colnames_side: str = "bottom",
    show_rownames: bool = True, show_colnames: bool = True,
    rownames_style: dict = dict(rotation=0, size=6),
    colnames_style: dict = dict(rotation=0, size=6),
    edgecolor: str = "none", edgewidth: float = 1,
    annotation_row: DataFrame = None, annotation_col: DataFrame = None,
    annotation_row_cmaps: Dict[str, Union[str, Colormap, list]] = None,
    annotation_col_cmaps: Dict[str, Union[str, Colormap, list]] = None,
    annotation_row_names_style: Dict = dict(size=6),
    annotation_col_names_style: Dict = dict(size=6),
    show_annotation_row_names: bool = True, show_annotation_col_names: bool = True,
    legend_tick_locs: Dict[str, Sequence] = None, legend_tick_labels: Dict[str, Sequence] = None,
    legend_tick_labels_styles: Dict = dict(size=6),
    legend_titles: Dict[str, bool] = None, legend_title_styles: Dict = dict(size=6),
    width: float = 8, height: float = 6, wspace: float = 0.1, hspace: float = 0.1,
    annotation_bar_width: float = 0.03, legend_bar_width: float = 1.5 * 0.03,
    annotation_bar_space: float = 0.2, legend_bar_space: float = 1
) -> Figure:
    """Plot heatmap with annotation bars

    Parameters
    ----------
    mat : DataFrame
        the main heatmap DataFrame
    cmap : Union[str, Colormap, list], optional
        the colormap of heatmap, by default "bwr"
    vmin : float, optional
        the minemum value scaled. by default None, use the minemum value of `mat`
    vmax : float, optional
        the maximum value scaled. by default None, use the maximum value of `mat`
    name : str, optional
        the name of heatmap, by default None, use "heatmap" as the name of heatmap
    rownames : ndarray, optional
        the row names provided, by default None, use the row names of main heatmap DataFrame
    colnames : ndarray, optional
        the column names provided, by default None, use the column names of main heatmap DataFrame
    rownames_side : str, optional
        show row names on which side("left" or "right")? by default "left"
    colnames_side : str, optional
        show column names on which side("top" or "bottom")? by default "bottom"
    show_rownames : bool, optional
        whether show row names? by default True
    show_colnames : bool, optional
        whether show column names? by default True
    rownames_style : dict, optional
        modify row names' style, such as, fontsize, fontstyle, etc. See more information on 
        [`matplotlib.text.Text`](https://matplotlib.org/stable/api/text_api.html#matplotlib.text.Text).
        by default dict(rotation=0, size=6)
    colnames_style : dict, optional
        see `rownames_style`, by default dict(rotation=0, size=6)
    edgecolor : str, optional
        the color of heatmap's cell edge, by default "none", no edge. 
        !Note: If provide `None`, will use the `rcParams["patch.edgecolor"]`, it default as "black".
    edgewidth : float, optional
        the width of heatmap's cell edge, by default 1
    annotation_row : DataFrame, optional
        DataFrame used to create row Annotationbar, by default None
    annotation_col : DataFrame, optional
        DataFrame used to create column Annotationbar, by default None
    annotation_row_cmaps : Dict[str, Union[str, Colormap, list]], optional
        Colormaps for each Annotationbar, keys are the DataFrame's columns, by default None, use
        "viridis" for continuous and "tab20" for discrete
    annotation_col_cmaps : Dict[str, Union[str, Colormap, list]], optional
        see `annotation_row_cmaps`, by default None
    annotation_row_names_style: Dict[str, Dict], optional
        modify the style of row AnnotationBar's name. Keys are the DataFrame's columns, by default
        None
    annotation_col_names_style: Dict[str, Dict], optional
        modify the style of column AnnotationBar's name. Keys are the DataFrame's columns, by
        default None
    show_annotation_row_names : bool, optional
        whether show row Annotationbar's name, by default True
    show_annotation_col_names : bool, optional
        whether show column Annotationbar's name, by default True
    legend_tick_locs : Dict[str, Sequence], optional
        modify the tick locations of legend, keys are the name of heatmap or the column names of
        annotation DataFrame, by default None
    legend_tick_labels : Dict[str, Sequence], optional
        modify the tick labels of legend, keys are the name of heatmap or the column names of
        annotation DataFrame. The length of each legend tick labels must be matched with its tick
        locations. by default None
    legend_tick_labels_styles : Dict, optional
        modify the each legend tick labels' style. Keys are the name of heatmap or the column names
        of annotation DataFrame. More informations see `rownames_style`. by default dict(size=6)
    legend_titles : Dict[str, bool], optional
        modify the each legend title. Others are the same as `legend_tick_labels`.
        by default None
    legend_title_styles : Dict, optional
        modify the each legend title's style. Others are the same as `legend_tick_labels_styles`.
        by default dict(size=6)
    width : float, optional
        the whole figure width, by default 8
    height : float, optional
        the whole figure height, by default 6
    wspace : float, optional
        the space of whole figure at the width direction, by default 0.1. It's the fraction of the
        average length of width's axises
    hspace : float, optional
        see wspace, by default 0.1
    annotation_bar_width : float, optional
        the width of Annotationbar, by default 0.03. It's the fraction of the whole figure width
    legend_bar_width : float, optional
        the width of legend bar, by default 1.5*0.03. It's the fraction of the whole figure width.
        And it's better, set it as 1.5 * annotation_bar_width.
    annotation_bar_space : float, optional
        the space between Annotationbars, by default 0.2. It's the fraction of the real
        Annotationbar width.
    legend_bar_space : float, optional
        the space between legend bars, by default 1. It's the fraction of the real legend bar width

    Returns
    -------
    Figure
    """
    # Heatmap
    # Check arguments
    rownames = check_margin_names(mat.index.to_numpy(), rownames, show_rownames, axis="row")
    colnames = check_margin_names(mat.columns.to_numpy(), colnames, show_colnames, axis="col")
    mat = mat.to_numpy()
    name = name if name is not None else "heatmap"

    # Instance class
    heatmap = Heatmap(
        mat=mat, cmap=cmap, vmin=vmin, vmax=vmax, name=name,
        rownames=rownames, colnames=colnames,
        rownames_side=rownames_side, colnames_side=colnames_side,
        rownames_style=rownames_style, colnames_style=colnames_style,
        edgecolor=edgecolor, edgewidth=edgewidth
    )

    # Row/Column Annotations
    row_annotationbars = create_annotation(
        anno=annotation_row, cmaps=annotation_row_cmaps, show_names=show_annotation_row_names,
        expected_nrows=heatmap.nrows, axis="row", names_style = annotation_row_names_style
    )
    col_annotationbars = create_annotation(
        anno=annotation_col, cmaps=annotation_col_cmaps, show_names=show_annotation_col_names,
        expected_nrows=heatmap.ncols, axis="col", names_style = annotation_col_names_style
    )

    # Legends
    legends = []
    legend_tick_locs = none2dict(legend_tick_locs)
    legend_tick_labels = none2dict(legend_tick_labels)
    legend_tick_labels_styles = none2dict(legend_tick_labels_styles)
    legend_titles = none2dict(legend_titles)
    legend_title_styles = none2dict(legend_title_styles)

    # Heatmap's legend
    legends.append(Legend(
        cmap=heatmap.cmap, norm=heatmap.norm, name=legend_titles.pop(heatmap.name, heatmap.name),
        tick_locs=legend_tick_locs.pop(
            heatmap.name, np.linspace(heatmap.norm.vmin, heatmap.norm.vmax, 5)),
        tick_labels=legend_tick_labels.pop(
            heatmap.name, np.linspace(heatmap.norm.vmin, heatmap.norm.vmax, 5)),
        tick_labels_params=legend_tick_labels_styles,
        title_params=legend_title_styles,
        bartype=CONTINUOUS
    ))

    # AnnotationBars legends
    if row_annotationbars is not None:
        for anno_bar in row_annotationbars.annotationbars:
            if anno_bar.bartype == CONTINUOUS:
                tick_locs = legend_tick_locs.pop(
                    anno_bar.name, np.linspace(anno_bar.norm.vmin, anno_bar.norm.vmax, 5))
                tick_labels = legend_tick_labels.pop(
                    anno_bar.name, np.linspace(anno_bar.norm.vmin, anno_bar.norm.vmax, 5))
            else:
                tick_locs = legend_tick_locs.pop(
                    anno_bar.name, list(anno_bar.values_mapper.values()))
                tick_labels = legend_tick_labels.pop(
                    anno_bar.name, list(anno_bar.values_mapper.keys()))

            legends.append(Legend(
                cmap=anno_bar.cmap, norm=anno_bar.norm, name=anno_bar.name,
                tick_locs=tick_locs, tick_labels=tick_labels,
                tick_labels_params=legend_tick_labels_styles,
                title_params=legend_title_styles,
                bartype=anno_bar.bartype
            ))
    if col_annotationbars is not None:
        for anno_bar in col_annotationbars.annotationbars:
            if anno_bar.bartype == CONTINUOUS:
                tick_locs = legend_tick_locs.pop(
                    anno_bar.name, np.linspace(anno_bar.norm.vmin, anno_bar.norm.vmax, 5))
                tick_labels = legend_tick_labels.pop(
                    anno_bar.name, np.linspace(anno_bar.norm.vmin, anno_bar.norm.vmax, 5))
            else:
                tick_locs = legend_tick_locs.pop(
                    anno_bar.name, list(anno_bar.values_mapper.values()))
                tick_labels = legend_tick_labels.pop(
                    anno_bar.name, list(anno_bar.values_mapper.keys()))

            legends.append(Legend(
                cmap=anno_bar.cmap, norm=anno_bar.norm, name=anno_bar.name,
                tick_locs=tick_locs, tick_labels=tick_labels,
                tick_labels_params=legend_tick_labels_styles,
                title_params=legend_title_styles,
                bartype=anno_bar.bartype
            ))

    n_leftbars = len(row_annotationbars.annotationbars) if row_annotationbars is not None else 1
    n_rightbars = len(legends) if len(legends) > 0 else 1
    n_topbars = len(col_annotationbars.annotationbars) if col_annotationbars is not None else 1
    n_bottombars = 1

    left_width = annotation_bar_width * width * n_leftbars + \
        annotation_bar_space * (annotation_bar_width * width) * (n_leftbars - 1)
    right_width = legend_bar_width * width * n_rightbars + \
        legend_bar_space * (legend_bar_width * width) * (n_rightbars - 1)
    top_height = (annotation_bar_width * width) * n_topbars + annotation_bar_space * \
        (annotation_bar_width * width) * (n_topbars - 1)
    bottom_height = (annotation_bar_width * width) * n_bottombars + annotation_bar_space * \
        (annotation_bar_width * width) * (n_bottombars - 1)

    center_width = width - left_width - right_width
    center_height = height - top_height - bottom_height
    layout = Layout(
        center_width=center_width, center_height=center_height,
        left_width=left_width, top_height=top_height,
        right_width=right_width, bottom_height=bottom_height,
        sub_left_width=[1] * n_leftbars, sub_top_height=[1] * n_topbars,
        sub_right_width=[1] * n_rightbars, sub_bottom_height=[1] * n_bottombars,
        wspace=wspace, hspace=hspace,
        sub_left_wspace=annotation_bar_space, sub_top_hspace=annotation_bar_space,
        sub_right_wspace=legend_bar_space, sub_bottom_hspace=annotation_bar_space,
        width=width, height=height
    )

    # Draw plots
    # Heatmap
    ht_ax = layout.create_axes(layout.gs[1, 1])
    heatmap.draw(ht_ax)

    # Annotation Bars
    if row_annotationbars is not None:
        row_annobars_axes = layout.create_axes(layout.left_gs, axis=1)
        for ax, annobar in zip(row_annobars_axes, row_annotationbars.annotationbars):
            annobar.draw(ax)
    if col_annotationbars is not None:
        col_annobars_axes = layout.create_axes(layout.top_gs, axis=0)
        for ax, annobar in zip(col_annobars_axes, col_annotationbars.annotationbars):
            annobar.draw(ax)
    if len(legends) > 0:
        legend_bars_axes = layout.create_axes(layout.right_gs, axis=1)
        for ax, legend in zip(legend_bars_axes, legends):
            legend.draw(ax)

    return layout.fig



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="pheatmap",  # Required
    version="1.0.1",  # Required
    description="pheatmap for Python",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/Ann-Holmes/pheatmap",  # Optional
    author="Zongliang Hou",  # Optional
    author_email="zonglianghou@163.com",  # Optional
    classifiers=[  # Optional
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: POSIX :: Linux"
    ],
    keywords="heatmap, gene, expression, visualization",  # Optional
    package_dir={"": "src"},  # Optional
    packages=find_packages(where="src", exclude=["tests"]),  # Required
    python_requires=">=3.8, <4",
    install_requires=["numpy", "matplotlib", "pandas"],  # Optional
    extras_require={  # Optional
        "dev": ["sphinx", "myst-parser"]
    },
    project_urls={  # Optional
        "Documents": "https://pheatmap.readthedocs.io/en/latest/",
        "Bug Reports": "https://github.com/Ann-Holmes/pheatmap/issues",
        "Source": "https://github.com/Ann-Holmes/pheatmap",
    },
)
